@@ -1,5 +1,5 @@
 """
-PyMule color module for color creation and conversion operations.
+color module for color creation and conversion operations.
 """
 from math import floor
 from .compat import *
@@ -10,26 +10,65 @@ __all__ = ["Color", "is_rgb_color", "is_rgba_color", "argb_to_color", "ARGB",
            "COLOR"]
 
 
+def _clip(val, _min, _max):
+    return max(min(val, _max), _min)
+
+
 class Color(object):
-    """A simple RGBA-based color implementation."""
+    """A class for working with and converting RGBA colors.
+    
+    This class represents the 4 RGBA color channels (red, green, blue, and
+    alpha transparency) as integers from 0 to 255. It also provides methods
+    for converting colors to alternative color spaces (e.g. HSV or CMY).
+
+    ``Color`` objects support basic arithmetic operations (``+, -, *, /, %``),
+    which operate on a per-channel basis. For example, the operation ::
+
+       color = color1 + color2
+
+    is the same as ::
+
+       color = Color()
+       color.r = min(color1.r + color2.r, 255)
+       color.g = min(color1.g + color2.g, 255)
+
+    All arithmetic operations guarantee that the channel values stay within
+    the allowed range of [0, 255].
+
+    Args:
+        r (int, optional): An integer between 0 and 255 indicating the red
+            level of the color. Defaults to 255.
+        g (int, optional): An integer between 0 and 255 indicating the green
+            level of the color. Defaults to 255.
+        b (int, optional): An integer between 0 and 255 indicating the blue
+            level of the color. Defaults to 255.
+        a (int, optional): An integer between 0 and 255 indicating the alpha
+            trasparency level of the color, with 0 being fully transparent and
+            255 being fully opaque. Defaults to 255.
+
+    """
+    # NOTE: should this use floats internally for better precision?
     def __init__(self, r=255, g=255, b=255, a=255):
-        """Creates a Color with the specified RGBA values."""
-        if r < 0 or r > 255:
-            raise ValueError("r must be in the range [0; 255]")
-        if g < 0 or g > 255:
-            raise ValueError("r must be in the range [0; 255]")
-        if b < 0 or b > 255:
-            raise ValueError("r must be in the range [0; 255]")
-        if a < 0 or a > 255:
-            raise ValueError("r must be in the range [0; 255]")
-        self._r = int(r)
-        self._g = int(g)
-        self._b = int(b)
-        self._a = int(a)
+        for val in (r, g, b, a):
+            self._verify_rgba_value(val)
+        self._r = float(int(r))
+        self._g = float(int(g))
+        self._b = float(int(b))
+        self._a = float(int(a))
+
+    def _verify_rgba_value(self, val):
+        """Verifies that the input is a valid uint8 RGBA value."""
+        e = "All RGBA color values must be integers between 0 and 255 (got {0})"
+        try:
+            float(val)
+        except (ValueError, TypeError):
+            raise TypeError(e.format(val))
+        if val < 0 or val > 255:
+            raise ValueError(e.format(val))
 
     def __repr__(self):
-        return "Color(r=%d, g=%d, b=%d, a=%d)" % \
-            (self.r, self.g, self.b, self.a)
+        cols = [self.r, self.g, self.b, self.a]
+        return "Color(r={0}, g={1}, b={2}, a={3})".format(*cols)
 
     def __copy__(self):
         return Color(self.r, self.g, self.b, self.a)
@@ -73,31 +112,35 @@ class Color(object):
 
     def __div__(self, color):
         vals = [0, 0, 0, 0]
-        if color.r != 0:
-            vals[0] = self.r / color.r
-        if color.g != 0:
-            vals[1] = self.g / color.g
-        if color.b != 0:
-            vals[2] = self.b / color.b
-        if color.a != 0:
-            vals[3] = self.a / color.a
+        if color._r != 0:
+            vals[0] = (self._r / color._r)
+        if color._g != 0:
+            vals[1] = (self._g / color._g)
+        if color._b != 0:
+            vals[2] = (self._b / color._b)
+        if color._a != 0:
+            vals[3] = (self._a / color._a)
         return Color(vals[0], vals[1], vals[2], vals[3])
 
     def __truediv__(self, color):
         vals = [0, 0, 0, 0]
-        if color.r != 0:
-            vals[0] = self.r / color.r
-        if color.g != 0:
-            vals[1] = self.g / color.g
-        if color.b != 0:
-            vals[2] = self.b / color.b
-        if color.a != 0:
-            vals[3] = self.a / color.a
+        if color._r != 0:
+            vals[0] = (self._r / color._r)
+        if color._g != 0:
+            vals[1] = (self._g / color._g)
+        if color._b != 0:
+            vals[2] = (self._b / color._b)
+        if color._a != 0:
+            vals[3] = (self._a / color._a)
         return Color(vals[0], vals[1], vals[2], vals[3])
 
     def __mul__(self, color):
-        vals = (min(self.r * color.r, 255), min(self.g * color.g, 255),
-                min(self.b * color.b, 255), min(self.a * color.a, 255))
+        vals = (
+            min(self._r * color._r, 255),
+            min(self._g * color._g, 255),
+            min(self._b * color._b, 255),
+            min(self._a * color._a, 255)
+        )
         return Color(vals[0], vals[1], vals[2], vals[3])
 
     def __sub__(self, color):
@@ -126,67 +169,61 @@ class Color(object):
 
     @property
     def r(self):
-        """Gets or sets the red value of the Color."""
-        return self._r
+        """"int: The 8-bit RGBA red level for the color."""
+        return int(round(self._r))
 
     @r.setter
     def r(self, val):
-        """Gets or sets the red value of the Color."""
-        if type(val) not in(int, long):
-            raise TypeError("value must be an int")
-        if val < 0 or val > 255:
-            raise ValueError("The value must be in the range [0; 255]")
-        self._r = val
+        self._verify_rgba_value(val)
+        self._r = float(val)
 
     @property
     def g(self):
-        """Gets or sets the green value of the Color."""
-        return self._g
+        """"int: The 8-bit RGBA green level for the color."""
+        return int(round(self._g))
 
     @g.setter
     def g(self, val):
-        """Gets or sets the green value of the Color."""
-        if type(val) not in(int, long):
-            raise TypeError("value must be an int")
-        if val < 0 or val > 255:
-            raise ValueError("The value must be in the range [0; 255]")
-        self._g = val
+        self._verify_rgba_value(val)
+        self._g = float(val)
 
     @property
     def b(self):
-        """Gets or sets the blue value of the Color."""
-        return self._b
+        """"int: The 8-bit RGBA blue level for the color."""
+        return int(round(self._b))
 
     @b.setter
     def b(self, val):
-        """Gets or sets the blue value of the Color."""
-        if type(val) not in(int, long):
-            raise TypeError("value must be an int")
-        if val < 0 or val > 255:
-            raise ValueError("The value must be in the range [0; 255]")
-        self._b = val
+        self._verify_rgba_value(val)
+        self._b = float(val)
 
     @property
     def a(self):
-        """Gets or sets the alpha value of the Color."""
-        return self._a
+        """"int: The 8-bit RGBA alpha transparency level for the color."""
+        return int(round(self._a))
 
     @a.setter
     def a(self, val):
-        """Gets or sets the alpha value of the Color."""
-        if type(val) not in(int, long):
-            raise TypeError("value must be an int")
-        if val < 0 or val > 255:
-            raise ValueError("The value must be in the range [0; 255]")
-        self._a = val
+        self._verify_rgba_value(val)
+        self._a = float(val)
 
     @property
     def hsva(self):
-        """The Color as HSVA value."""
-        rn = self.r / 255.0
-        gn = self.g / 255.0
-        bn = self.b / 255.0
-        an = self.a / 255.0
+        """tuple: A representation of the color in HSV(A) color space.
+        
+        The HSVA color space represents colors in terms of hue, saturation,
+        value (brightness), and alpha (transparency). Hue is represented as a
+        value on color wheel between 0 and 360, whereas saturation, brightness,
+        and alpha are all represented as values from 0 to 100.
+
+        Note that due to rounding errors, this may not return the exact HSVA
+        values for the given color.
+        
+        """
+        rn = self._r / 255.0
+        gn = self._g / 255.0
+        bn = self._b / 255.0
+        an = self._a / 255.0
 
         maxv = max(rn, gn, bn)
         minv = min(rn, gn, bn)
@@ -209,11 +246,11 @@ class Color(object):
             h = (60 * (rn - gn) / diff) + 240.0
         if h < 0:
             h += 360.0
+
         return (h, s, v, a)
 
     @hsva.setter
     def hsva(self, value):
-        """The Color as HSVA value."""
         h, s, v, a = value
         for x in (h, s, v, a):
             if type(x) not in(int, long, float):
@@ -222,49 +259,45 @@ class Color(object):
                 not (0 <= a <= 100) or not (0 <= h <= 360):
             raise ValueError("invalid HSVA value")
 
+        hi = int(floor(h / 60.0))
+        if hi > 5:
+            raise OverflowError("invalid HSVA value")
+
         self.a = int((a / 100.0) * 255)
         s /= 100.0
         v /= 100.0
-
-        hi = int(floor(h / 60.0))
         f = (h / 60.0) - hi
         p = v * (1 - s)
         q = v * (1 - s * f)
         t = v * (1 - s * (1 - f))
-        if hi == 0:
-            self.r = int(v * 255)
-            self.g = int(t * 255)
-            self.b = int(p * 255)
-        elif hi == 1:
-            self.r = int(q * 255)
-            self.g = int(v * 255)
-            self.b = int(p * 255)
-        elif hi == 2:
-            self.r = int(p * 255)
-            self.g = int(v * 255)
-            self.b = int(t * 255)
-        elif hi == 3:
-            self.r = int(p * 255)
-            self.g = int(q * 255)
-            self.b = int(v * 255)
-        elif hi == 4:
-            self.r = int(t * 255)
-            self.g = int(p * 255)
-            self.b = int(v * 255)
-        elif hi == 5:
-            self.r = int(v * 255)
-            self.g = int(p * 255)
-            self.b = int(q * 255)
-        else:
-            raise OverflowError("invalid HSVA value")
+        rgb_map = [
+            (v, t, p), # if hi == 0
+            (q, v, p), # if hi == 1
+            (p, v, t), # if hi == 2
+            (p, q, v), # if hi == 3
+            (t, p, v), # if hi == 4
+            (v, p, q)  # if hi == 5
+        ]
+        vals = [_clip(n * 255.0, 0.0, 255.0) for n in rgb_map[hi]]
+        self.r, self.g, self.b = vals
 
     @property
     def hsla(self):
-        """The Color a HSLA value."""
-        rn = self.r / 255.0
-        gn = self.g / 255.0
-        bn = self.b / 255.0
-        an = self.a / 255.0
+        """tuple: A representation of the color in HSL(A) color space.
+        
+        The HSLA color space represents colors in terms of hue, saturation,
+        lightness, and alpha (transparency). Hue is represented as a
+        value on color wheel between 0 and 360, whereas saturation, lightness,
+        and alpha are all represented as values from 0 to 100.
+
+        Note that due to rounding errors, this may not return the exact HSLA
+        values for the given color.
+        
+        """
+        rn = self._r / 255.0
+        gn = self._g / 255.0
+        bn = self._b / 255.0
+        an = self._a / 255.0
 
         maxv = max(rn, gn, bn)
         minv = min(rn, gn, bn)
@@ -276,7 +309,7 @@ class Color(object):
         a = an * 100.0
 
         if maxv == minv:
-            return(h, s, l, a)
+            return (h, s, l, a)
 
         if l <= 50.0:
             s = diff / (maxv + minv) * 100.0
@@ -291,11 +324,11 @@ class Color(object):
             h = (60 * (rn - gn) / diff) + 240.0
         if h < 0:
             h += 360.0
+
         return (h, s, l, a)
 
     @hsla.setter
     def hsla(self, value):
-        """The Color a HSLA value."""
         h, s, l, a = value
         for x in (h, s, l, a):
             if type(x) not in (int, long, float):
@@ -305,14 +338,12 @@ class Color(object):
             raise ValueError("invalid HSLA value")
 
         self.a = int((a / 100.0) * 255)
-
         s /= 100.0
         l /= 100.0
-
         if s == 0:
-            self.r = int(l * 255)
-            self.g = int(l * 255)
-            self.b = int(l * 255)
+            self.r = l * 255.0
+            self.g = l * 255.0
+            self.b = l * 255.0
             return
 
         q = 0
@@ -323,71 +354,51 @@ class Color(object):
         p = 2 * l - q
 
         ht = h / 360.0
+        vals = []
+        for h in [ht + (1.0 / 3.0), ht, ht - (1.0 / 3.0)]:
+            if h < 0:
+                h += 1
+            elif h > 1:
+                h -= 1
+            if h < (1.0 / 6.0):
+                val = (p + ((q - p) * 6 * h))
+            elif h < 0.5:
+                val = q
+            elif h < (2.0 / 3.0):
+                val = (p + ((q - p) * 6 * (2.0 / 3.0 - h)))
+            else:
+                val = p
+            vals.append(_clip(val * 255.0, 0.0, 255.0))
 
-        # r
-        h = ht + (1.0 / 3.0)
-        if h < 0:
-            h += 1
-        elif h > 1:
-            h -= 1
-
-        if h < (1.0 / 6.0):
-            self.r = int((p + ((q - p) * 6 * h)) * 255)
-        elif h < 0.5:
-            self.r = int(q * 255)
-        elif h < (2.0 / 3.0):
-            self.r = int((p + ((q - p) * 6 * (2.0 / 3.0 - h))) * 255)
-        else:
-            self.r = int(p * 255)
-
-        # g
-        h = ht
-        if h < 0:
-            h += 1
-        elif h > 1:
-            h -= 1
-
-        if h < (1.0 / 6.0):
-            self.g = int((p + ((q - p) * 6 * h)) * 255)
-        elif h < 0.5:
-            self.g = int(q * 255)
-        elif h < (2.0 / 3.0):
-            self.g = int((p + ((q - p) * 6 * (2.0 / 3.0 - h))) * 255)
-        else:
-            self.g = int(p * 255)
-
-        # b
-        h = ht - (1.0 / 3.0)
-        if h < 0:
-            h += 1
-        elif h > 1:
-            h -= 1
-
-        if h < (1.0 / 6.0):
-            self.b = int((p + ((q - p) * 6 * h)) * 255)
-        elif h < 0.5:
-            self.b = int(q * 255)
-        elif h < (2.0 / 3.0):
-            self.b = int((p + ((q - p) * 6 * (2.0 / 3.0 - h))) * 255)
-        else:
-            self.b = int(p * 255)
+        self.r, self.g, self.b = vals
 
     @property
     def i1i2i3(self):
-        """The Color as I1I2I3 value."""
-        rn = self.r / 255.0
-        gn = self.g / 255.0
-        bn = self.b / 255.0
+        """tuple: A representation of the color in I1I2I3 color space.
+        
+        The I1I2I3 color space represents colors in terms of a color-independent
+        intensity level (I1) and two chromatic channels (I2 and I3), with the
+        aim of minimizing correlations between its channels for natural images.
+        Intensity (I1) is represented as a float between 0.0 and 1.0, whereas
+        the color channels (I2 and I3) are represented as floats between
+        -0.5 and 0.5, inclusive.
 
-        i1 = (rn + gn + bn) / 3.0
-        i2 = (rn - bn) / 2.0
-        i3 = (2 * gn - rn - bn) / 4.0
+        Note that due to rounding errors, this may not return the exact I1I2I3
+        values for the given color.
+        
+        """
+        rn = self._r / 255.0
+        gn = self._g / 255.0
+        bn = self._b / 255.0
 
-        return(i1, i2, i3)
+        i1 = _clip((rn + gn + bn) / 3.0, 0.0, 1.0)
+        i2 = _clip((rn - bn) / 2.0, -0.5, 0.5)
+        i3 = _clip((2 * gn - rn - bn) / 4.0, -0.5, 0.5)
+
+        return (i1, i2, i3)
 
     @i1i2i3.setter
     def i1i2i3(self, value):
-        """The Color as I1I2I3 value."""
         i1, i2, i3 = value
         for x in (i1, i2, i3):
             if type(x) not in (int, long, float):
@@ -400,37 +411,52 @@ class Color(object):
         ar = 2 * i2 + ab
         ag = 3 * i1 - ar - ab
 
-        self.r = int(ar * 255)
-        self.g = int(ag * 255)
-        self.b = int(ab * 255)
+        self.r = _clip(ar * 255, 0.0, 255.0)
+        self.g = _clip(ag * 255, 0.0, 255.0)
+        self.b = _clip(ab * 255, 0.0, 255.0)
 
     @property
     def cmy(self):
-        """The Color as CMY value."""
-        return (1.0 - self.r / 255.0,
-                1.0 - self.g / 255.0,
-                1.0 - self.b / 255.0)
+        """tuple: A representation of the color in CMY color space.
+
+        The CMY color space is the inverse of the RGB color space, and
+        represents colors in subtractive amounts of Cyan, Magenta, and Yellow.
+        All three values are represented as floats between 0.0 and 1.0.
+
+        """
+        return (1.0 - self._r / 255.0,
+                1.0 - self._g / 255.0,
+                1.0 - self._b / 255.0)
 
     @cmy.setter
     def cmy(self, value):
-        """The Color as CMY value."""
         c, m, y = value
         if (c < 0 or c > 1) or (m < 0 or m > 1) or (y < 0 or y > 1):
             raise ValueError("invalid CMY value")
-        self.r = int((1.0 - c) * 255)
-        self.g = int((1.0 - m) * 255)
-        self.b = int((1.0 - y) * 255)
+        self._r = (1.0 - c) * 255
+        self._g = (1.0 - m) * 255
+        self._b = (1.0 - y) * 255
 
     def normalize(self):
-        """Returns the RGBA values in a normalized form with the range
-        [0;1] as tuple.
+        """Returns the RGBA values as floats between 0 and 1.
+
+        Returns:
+            tuple: The (r, g, b, a) values of the color as normalized floats.
+
         """
         return (self.r / 255.0, self.g / 255.0, self.b / 255.0, self.a / 255.0)
 
 
 def is_rgb_color(v):
-    """Checks, if the passed value is an item that could be converted to
-    a RGB color.
+    """Checks whether a value be converted to an RGB color.
+
+    Args:
+        v: The value to try and interpret as an RGB color.
+    
+    Returns:
+        bool: True if the value can be interpreted as an RGB color, otherwise
+            False.
+
     """
     try:
         if hasattr(v, "r") and hasattr(v, "g") and hasattr(v, "b"):
@@ -447,9 +473,17 @@ def is_rgb_color(v):
         return False
 
 
+# NOTE: Add support for trying to parse strs and ints using below functions?
 def is_rgba_color(v):
-    """Checks, if the passed value is an item that could be converted to
-    a RGBA color.
+    """Checks whether a value be converted to an RGBA color.
+
+    Args:
+        v: The value to try and interpret as an RGBA color.
+    
+    Returns:
+        bool: True if the value can be interpreted as an RGBA color, otherwise
+            False.
+
     """
     rgb = is_rgb_color(v)
     if not rgb:
@@ -465,9 +499,16 @@ def is_rgba_color(v):
         return False
 
 
+# TODO: Add type-checking/exceptions?
 def argb_to_color(v):
-    """Converts an integer value to a Color, assuming the integer
-    represents a 32-bit ARGB value.
+    """Converts a 32-bit ARGB integer value to a :obj:`sdl2.ext.Color`.
+
+    Args:
+        v (int): An integer representing a color in ARGB format.
+
+    Returns:
+        :obj:`sdl2.ext.Color`: An object representing the given color.
+
     """
     v = long(v)
 
@@ -482,8 +523,14 @@ ARGB = argb_to_color
 
 
 def rgba_to_color(v):
-    """Converts an integer value to a Color, assuming the integer
-    represents a 32-bit RGBBA value.
+    """Converts a 32-bit RGBA integer value to a :obj:`sdl2.ext.Color`.
+
+    Args:
+        v (int): An integer representing a color in RGBA format.
+
+    Returns:
+        :obj:`sdl2.ext.Color`: An object representing the given color.
+
     """
     v = long(v)
 
@@ -498,19 +545,25 @@ RGBA = rgba_to_color
 
 
 def string_to_color(s):
-    """Converts a hex color string or color name to a Color value.
+    """Converts a hex color string to a Color value.
 
-    Supported hex values are:
+    Hex colors can be specified in any of the following formats:
 
-    #RGB
-    #RGBA
-    #RRGGBB
-    #RRGGBBAA
+    * #RGB
+    * #RGBA
+    * #RRGGBB
+    * #RRGGBBAA
+    * 0xRGB
+    * 0xRGBA
+    * 0xRRGGBB
+    * 0xRRGGBBAA
+
+    Args:
+        s (str): A valid hex color in string format.
+
+    Returns:
+        :obj:`sdl2.ext.Color`: An object representing the given color.
 
-    0xRGB
-    0xRGBA
-    0xRRGGBB
-    0xRRGGBBAA
     """
     if type(s) is not str:
         raise TypeError("s must be a string")
@@ -543,9 +596,19 @@ def string_to_color(s):
 
 
 def convert_to_color(v):
-    """Tries to convert the passed value to a Color object.
+    """Tries to convert an arbitrary object to a :obj:`sdl2.ext.Color`.
+
+    If an integer is provided, it is assumed to be in ARGB layout.
+
+    Args:
+        v: An arbitrary object type representing a color.
+
+    Returns:
+        :obj:`sdl2.ext.Color`: An object representing the given color.
+
+    Raises:
+        ValueError: If the value could not be converted successfully.
 
-    If the color is an integer value, it is assumed to be in ARGB layout.
     """
     if isinstance(v, Color):
         return v
